import os
import socket
import subprocess

def get_os_release(args, host=None):
    freebsd = False
    out = None
    cmdarr = ["/bin/cat /etc/os-release"]
    if host and host['remote']:
        out = run_remote_ssh_command(args, host, cmdarr[0])
    else:
        try:
            out = subprocess.check_output(cmdarr, shell=True)
        except subprocess.CalledProcessError:
            logging.error("Error running local command")

    if out is None or out.strip() == '':
        # try FreeBSD
        cmdarr = ["/usr/bin/uname -v -p"]
        if host and host['remote']:
            out = run_remote_ssh_command(args, host, cmdarr[0])
        else:
            try:
                out = subprocess.check_output(cmdarr, shell=True)
            except subprocess.CalledProcessError:
                logging.error("Error running local command")

        if out is not None and 'FreeBSD' not in out:
            # try OpenBSD
            cmdarr = ["/usr/bin/uname -srvm"]
            if host and host['remote']:
                out = run_remote_ssh_command(args, host, cmdarr[0])
            else:
                try:
                    out = subprocess.check_output(cmdarr, shell=True)
                except subprocess.CalledProcessError:
                    logging.error("Error running local command")

    if out is None:
        logging.error("Failed to get os-release")
        return None

    if 'FreeBSD' in out or 'OpenBSD' in out:
        return out
    else:
        output_lines = out.splitlines()
        for l in output_lines:
            if 'PRETTY_NAME' in l:
                return l.split('=')[1].replace('"','')
    return None


def get_asset_type(os):
    if "CentOS" in os:
        return "CentOS"
    elif "Red Hat" in os:
        return "Red Hat"
    elif "Ubuntu" in os:
        return "Ubuntu"
    elif "Debian" in os:
        return "Debian"
    elif "Amazon Linux" in os:
        return "Amazon Linux"
    elif "Oracle Linux" in os:
        return "Oracle Linux"
    elif "FreeBSD" in os:
        return "FreeBSD"
    elif "OpenBSD" in os:
        return "OpenBSD"
    else:
        logging.error('Not a supported OS type')
        return None


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

def find_files(localpath, filename):
    ret_files = []
    for root, subdirs, files in os.walk(localpath):
        for fname in files:
            file_path = os.path.join(root, fname)
            if len(filename) == 0:
                ret_files.append(file_path)
            elif file_path.endswith(filename):
                ret_files.append(file_path)
    return ret_files

def ascii_string(in_str):
    ascii_str = ''.join([c if ord(c) < 128 else ' ' for c in in_str.strip()])
    return ascii_str

def get_indent(line):
    return len(line) - len(line.lstrip(' '))



import sys
import platform
import os
import json
import logging
import requests

_builtin_policies = [
    "no_do_now_impacts",
    "no_strong_copylefts",
    "no_code_secrets"
]

_allowed_fields_by_type = { 
    'vulnerability': {'priority': ["do now", "do later"] },
    'license': {'copyleft level': []},
    'code_secret': {'status': [], 'regex':[]},
    'dast': {}
}

_allowed_operators = [ '=', '!=', '>', '<', '>=', '<=']

_supported_action_types = ["on pass", "on fail"]

_allowed_actions = ['exit with code']

def validate_policy_file(policy_json_file):
    if os.path.isfile(policy_json_file) == False:
        logging.error('Error sepcified file [%s] not found...', policy_json_file)
        sys.exit(1)

    with open(policy_json_file, "r") as fd:
        try:
            policy_json = json.load(fd)
        except json.JSONDecodeError as ex:
            logging.error('Decoding JSON file failed with error [%s]', ex)
            sys.exit(1)

    ret_val = True
    for policy in policy_json:
        temp = validate_policy(policy)
        if ret_val == True:
            ret_val = temp

    if ret_val == False:
        logging.error('One or more policy validations failed...exiting!')
        sys.exit(1)

    return policy_json

def validate_policy(policy):
    if policy.get('name') is None:
        logging.error('Policy is missing field [%s]', policy_field)
        return False
    policy_name = policy['name']
    logging.info('Validating policy [%s]', policy_name)

    ret_val = True

    using_builtin_policy = False
    global _builtin_policies
    builtin_policy_name = policy.get('builtin_policy')
    if builtin_policy_name is not None:
        if builtin_policy_name in _builtin_policies:
            using_builtin_policy = True
        else:
            logging.error("Error: Policy [%s] is using unsupported builtin policy [%s]", policy_name, builtin_policy_name)
            return False

    policy_fields = ['type', 'conditions', 'actions'] if not using_builtin_policy else ['actions']

    # check required fields exist in policy
    for policy_field in policy_fields:
        if policy.get(policy_field) is None:
            ret_val = False
            logging.error('Policy [%s] is missing field [%s]', policy_name, policy_field)

    if not using_builtin_policy:
        condition_fields = ['field', 'operator', 'value']
        for condition in policy['conditions']:
            for cf in condition_fields:
                if condition.get(cf) is None:
                    logging.error('Condition [%s] in policy [%s] is missing field [%s]', condition, policy_name, cf)
                    ret_val = False

    actions = policy.get('actions')
    if len(actions) == 0:
        logging.error('There are no actions specified in policy [%s]...', policy_name)
        ret_val = False

    if ret_val == False:
        return False

    global _allowed_fields_by_type

    if not using_builtin_policy:
        policy_type = policy['type'].lower()
        if policy_type not in _allowed_fields_by_type.keys():
            logging.error('Policy [%s] has invalid type [%s]', policy_name, policy_type)
            return False
        else:
            policy['type'] = policy_type

        global _allowed_operators

        for condition in policy['conditions']:
            if condition['field'].lower() in _allowed_fields_by_type[policy_type].keys():
                condition['field'] = condition['field'].lower()
            else:
                logging.error('Invalid field [%s] specified in condition [%s] in policy [%s]', condition['field'], condition, policy_name)
                return False
            if condition['operator'] not in _allowed_operators:
                logging.error('Invalid operator [%s] specified in condition [%s] in policy [%s]', condition['operator'], condition, policy_name)
                ret_val = False
            if len(allowed_fields_by_type[policy_type][condition['field'].lower()]) > 0:
                if condition['value'].lower() not in _allowed_fields_by_type[policy_type][condition['field'].lower()]:
                    logging.error('Invalid value [%s] specified in condtion [%s] in policy [%s]', condition['value'], condition, policy_name)
                    ret_val = False

    global _supported_action_types
    global _allowed_actions
    actions = policy['actions']
    for action_type in actions.keys():
        if action_type not in _supported_action_types:
            logging.error('Unsupported action category [%s]', action_type)
            ret_val = False
    for action_type in _supported_action_types:
        actions = policy['actions'].get(action_type)
        if actions is not None:
            for action in actions:
                if action.lower() not in _allowed_actions:
                    logging.error('Invalid action [%s] specified in policy [%s]', action, policy_name)
                    ret_val = False
    if ret_val:
        logging.info("Policy validated successfully...")
    return ret_val

def apply_policy(policy_json, asset_id_list, args):
    url = "https://" + args.instance + "/api/v1/policies/apply/"
    auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"
    payload = { "asset_ids": asset_id_list, "policy_json": policy_json }
    resp = requests.post(url + auth_data, json=payload)
    if resp.status_code == 200:
        logging.info("Applying specified policy....")
        policy_job_id = resp.json()['policy_job_id']
        logging.info("Policy job started...job id is [%s]. This may take sometime...", policy_job_id)
        return policy_job_id
    else:
        logging.error("Error applying specified policy")
        logging.error("Response details: %s", resp.content)
        sys.exit(1)

def is_policy_job_done(policy_job_id, args):
    url = "https://" + args.instance + "/api/v1/policyjobs/" + policy_job_id + "/"
    auth_data = "?handle=" + args.handle + "&token=" + args.token + "&format=json"
    resp = requests.get(url + auth_data)
    if resp.status_code == 200:
        policy_job_json = resp.json()
        if policy_job_json['status'] == "COMPLETED":
            return True, policy_job_json
        else:
            return False, policy_job_json
    else:
        logging.error("Error retrieving policy job details for [%s]", policy_job_id)
        logging.error("Response details: %s", resp.content)
        return False, None

def process_policy_job_actions(pj_json):
    exit_with_code = False
    exit_code = None
    final_actions = { }
    policies = pj_json['policy_json']
    policies_outcome = pj_json['policy_outcome']
    for policy in policies:
        pn = policy['name']
        pa = policy['actions']
        policy_outcome = policies_outcome[pn]
        for assetid in policy_outcome.keys():
            po = policy_outcome[assetid]
            if po == "PASSED":
                lookup = 'on pass'
            elif po == "FAILED":
                lookup = 'on fail'
            if final_actions.get(pn) is None:
                final_actions[pn] = { }
            if pa.get(lookup) is not None:
                final_actions[pn][assetid] = pa[lookup]
                if "exit with code" in pa[lookup].keys():
                    exit_with_code = True
                    exit_code = pa[lookup]['exit with code']

    # Process policies with actions other than 'exit_with_code' in final_actions
    # TODO later on

    logging.info("Policy evaluation is done...")
    if exit_with_code:
        return exit_code
    else:
        return None




import sys
import platform
import os
import subprocess
import logging
import socket
import csv
import ipaddress
import paramiko
import getpass
import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.fernet import Fernet
import socket
import utils

def run_remote_ssh_command(args, host, command):
    assetid = host['assetid'] if host.get('assetid') is not None else host['hostname']
    output = ''
    try:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy)
        if host.get('userpwd') is not None and len(host['userpwd']) > 0 and (host.get('privatekey') is None or len(host['privatekey'])==0):
            client.connect(host['hostname'],username=host['userlogin'],password=host['userpwd'])
        elif host.get('privatekey') is not None and len(host['privatekey']) > 0:
            if host.get('userpwd') is not None and len(host['userpwd']) > 0:
                client.connect(host['hostname'],username=host['userlogin'],key_filename=host['privatekey'],passphrase=host['userpwd'])
            else:
                client.connect(host['hostname'],username=host['userlogin'],key_filename=host['privatekey'])
        else:
            client.connect(host['hostname'],username=host['userlogin'])
        stdin, stdout, stderr = client.exec_command(command)
        for line in stdout:
            output = output + line
        client.close()
    except paramiko.ssh_exception.AuthenticationException as e:
        logging.info("Authentication failed for asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except paramiko.ssh_exception.SSHException as e:
        logging.info("SSHException while connecting to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except socket.error as e:
        logging.info("Socket error while connection to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except:
        logging.info("Unknown error running remote discovery for asset [%s], host [%s]: [%s]", assetid, host['hostname'], sys.exc_info()[0])
        output = None
    finally:
        return output

def check_host_up(host):
    if not host['remote']:
        return True
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(2)
    try:
        result = sock.connect_ex((host['hostname'], 22))
        if result == 0:
            return True
        else:
            return False
    except socket.timeout:
        logging.error("Socket timeout")
        return False

def discover_openbsd(args, host):
    plist = []
    cmdarr = ["/usr/sbin/pkg_info -A"]
    logging.info("Retrieving product details")
    if host['remote']:
        pkgout = run_remote_ssh_command(args, host, cmdarr[0])
        if pkgout is None:
            return None
    else:
        try:
            pkgout = subprocess.check_output(cmdarr, shell=True)
        except subprocess.CalledProcessError:
            logging.error("Error running inventory")
            return None 

    begin = False
    for l in pkgout.splitlines():
        lsplit = l.split()
        pkgline = lsplit[0]
        ldash = pkgline.rfind('-')
        pkg = pkgline[:ldash] + ' ' + pkgline[ldash + 1:]
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details")
    return plist


def discover_freebsd(args, host):
    plist = []
    cmdarr = ["/usr/sbin/pkg info"]
    logging.info("Retrieving product details")
    if host['remote']:
        pkgout = run_remote_ssh_command(args, host, cmdarr[0])
        if pkgout is None:
            return None
    else:
        try:
            pkgout = subprocess.check_output(cmdarr, shell=True)
        except subprocess.CalledProcessError:
            logging.error("Error running inventory")
            return None 

    begin = False
    for l in pkgout.splitlines():
        lsplit = l.split()
        pkgline = lsplit[0]
        ldash = pkgline.rfind('-')
        pkg = pkgline[:ldash] + ' ' + pkgline[ldash + 1:]
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details")
    return plist

def discover_rh(args, host):
    plist = []
    cmdarr = ["/usr/bin/yum list installed"]
    logging.info("Retrieving product details")
    if host['remote']:
        yumout = run_remote_ssh_command(args, host, cmdarr[0])
        if yumout is None:
            return None
    else:
        try:
            yumout = subprocess.check_output(cmdarr, shell=True)
        except subprocess.CalledProcessError:
            logging.error("Error running inventory")
            return None 

    begin = False
    for l in yumout.splitlines():
        if 'Installed Packages' in l:
            begin = True
            continue
        if not begin:
            continue
        lsplit = l.split()
        pkg = lsplit[0]
        if len(lsplit) > 1:
            ver = lsplit[1]
        else:
            ver = ''
        pkgsp = pkg.split(".")
        if len(pkgsp) > 1:
            pkg = pkgsp[0]
            arch = pkgsp[1]
        else:
            pkg = pkgsp[0]
            arch = "noarch"

        if ':' in ver:
            ver = ver.split(':')[1]
        ver = ver + "." + arch
        logging.debug("Found product [%s %s]", pkg, ver)
        plist.append(pkg+' '+ver)
    logging.info("Completed retrieval of product details")
    return plist

def discover_ubuntu(args, host):
    plist = []
    cmdarr = ["/usr/bin/apt list --installed"]
    logging.info("Retrieving product details")
    if host['remote']:
        yumout = run_remote_ssh_command(args, host, cmdarr[0])
        if yumout is None:
            return None
    else:
        try:
            yumout = subprocess.check_output(cmdarr, shell=True)
        except subprocess.CalledProcessError:
            logging.error("Error running inventory")
            return None 

    begin = False
    for l in yumout.splitlines():
        if 'Listing...' in l:
            begin = True
            continue
        if not begin:
            continue
        if l.strip() == '':
            continue
        lsplit = l.split()
        pkg = lsplit[0].split('/')[0]
        ver = lsplit[1]
        logging.debug("Found product [%s %s]", pkg, ver)
        plist.append(pkg+' '+ver)
    logging.info("Completed retrieval of product details")
    return plist

def discover(args):
    handle = args.handle
    token = args.token
    instance = args.instance

    host_list_file = args.remote_hosts_csv
    if host_list_file == None:
        host_list_file = args.host_list

    if host_list_file is not None:
        with open(host_list_file, mode='r') as csv_file:
            csv_reader = csv.DictReader(csv_file, quoting=csv.QUOTE_NONE, escapechar='\\')
            password = None
            remote_hosts = []
            for row in csv_reader:
                if not args.secure and row['userpwd'].startswith('__SECURE__:'):
                    if args.password is None:
                        if password is None:
                            password = getpass.getpass(prompt="Enter password: ")
                    else:
                        password = args.password
                    password = password.encode()
                    salt = base64.b64encode(password)
                    kdf = PBKDF2HMAC(
                            algorithm=hashes.SHA256(),
                            length=32,
                            salt=salt,
                            iterations=100000,
                            backend=default_backend())
                    key = base64.urlsafe_b64encode(kdf.derive(password))
                    f = Fernet(key)
                    try:
                        epass = row['userpwd'].replace('__SECURE__:','')
                        row['userpwd'] = f.decrypt(epass)
                    except:
                        logging.error("Failed to decrypt login details for "+row['hostname'])
                        return None 
                elif row['userpwd'] != '' and not args.secure:
                    logging.warning('Unsecure login information in file. Use --secure to encrypt.')
                if '-' in row['hostname'] or '/' in row['hostname']: # IP range or CIDR is specified, then expand it
                    if '-' in row['hostname']:
                        iprange = row['hostname']
                        iprange = iprange.replace(' ','')
                        tokens = iprange.split('-')
                        if len(tokens) != 2 or len(tokens[0])==0 or len(tokens[1])==0:
                            logging.error("Skipping invalid range [%s]", row['hostname'])
                            continue
                        logging.info("Enumerating IPs based on specified range [%s]", iprange)
                        try:
                            startip = ipaddress.IPv4Address(unicode(tokens[0]))
                            endip = ipaddress.IPv4Address(unicode(tokens[1]))
                            cidrs = []
                            cidrs = [ipaddr for ipaddr in ipaddress.summarize_address_range(startip,endip)]
                        except Exception as e:
                            logging.error("Encountered exception: %s",e)
                            logging.error("Error converting IP range [%s] to CIDRs. Skipping it...", iprange)
                            continue
                        logging.info("Converted IP range [%s] to CIDRs %s", iprange, cidrs)
                    if '/' in row['hostname']:
                        logging.info("Enumerating IPs based on specified CIDR [%s]", row['hostname'])
                        try:
                            cidrs = [ipaddress.ip_network(unicode(row['hostname'],"ascii"))]
                        except Exception as e:
                            logging.error("Encountered exception: %s",e)
                            logging.error("Invalid CIDR [%s] specified. Skipping it...", row['hostname'])
                            continue
                    for cidr in cidrs:
                        for a in cidr:
                            trow = row.copy()
                            trow['hostname'] = str(a)

                            # Remove hard-coded asset ID and name for CIDR, as it will overwrite same asset
                            # These will based on host IP address automatically
                            trow['assetid'] = None
                            trow['assetname'] = None
                        
                            remote_hosts.append(trow)
                            remote_hosts[-1]['remote'] = True
                            logging.info("Enumerated IP: %s", a)
                else:
                    remote_hosts.append(row)
                    remote_hosts[-1]['remote'] = True
        if args.secure:
            # secure the host list
            logging.info("Securing host list file")
            if args.password is None:
                pp1 = getpass.getpass(prompt="Enter password: ")
                pp2 = getpass.getpass(prompt="Re-enter password: ")
                if pp1 != pp2:
                    logging.info("Passwords don't match. Try again.")
                    return None
            else:
                pp1 = args.password
            password = pp1.encode()
            salt = base64.b64encode(password)
            kdf = PBKDF2HMAC(
                    algorithm=hashes.SHA256(),
                    length=32,
                    salt=salt,
                    iterations=100000,
                    backend=default_backend())
            key = base64.urlsafe_b64encode(kdf.derive(password))
            f = Fernet(key)
            # verify the key if possible
            with open(host_list_file, mode='r') as csv_file:
                csv_reader = csv.DictReader(csv_file, quoting=csv.QUOTE_NONE, escapechar='\\')
                for row in csv_reader:
                    try:
                        if row['userpwd'] != '' and row['userpwd'].startswith('__SECURE__:'):
                            epass = row['userpwd'].replace('__SECURE__:','')
                            epass  = f.decrypt(epass)
                    except:
                        logging.error("Invalid password")
                        logging.error("Please use the same password as was used previously to secure the file")
                        return None
            # secure the new rows in the file
            with open(host_list_file, mode='w') as csvfile:
                fieldnames = ['hostname','userlogin','userpwd','privatekey','assetid','assetname']
                writer = csv.DictWriter(csvfile, fieldnames=fieldnames, quoting=csv.QUOTE_NONE, escapechar='\\')
                writer.writeheader()
                for h in remote_hosts:
                    if h['userpwd'] != '' and not h['userpwd'].startswith('__SECURE__:'):
                        h['userpwd'] = '__SECURE__:'+f.encrypt(h['userpwd'])
                    del h['remote']
                    writer.writerow(h)
            logging.info("Host list file secured")
            return None
        else:
            return discover_hosts(args, remote_hosts)
    else:
        host = { }
        host['assetid'] = utils.get_ip() if args.assetid is None else args.assetid
        host['assetname'] = host['assetid'] if args.assetname is None else args.assetname
        host['remote'] = False
        hosts = [ host ]
        return discover_hosts(args, hosts)

def discover_hosts(args, hosts):

    assets = []
    for host in hosts:
        asset = discover_host(args, host)
        if asset is not None:
            assets.append(asset)
    return assets

def discover_host(args, host):

    asset_id = host['assetid'] if host.get('assetid') is not None and len(host['assetid']) > 0 else host['hostname']
    asset_name = host['assetname'] if host.get('assetname') is not None and len(host['assetname']) > 0 else asset_id

    asset_id = asset_id.replace('/','-')
    asset_id = asset_id.replace(':','-')
    asset_name = asset_name.replace('/','-')
    asset_name = asset_name.replace(':','-')

    logging.info("Checking if host [%s] is reachable", asset_id)
    if not check_host_up(host):
        logging.error("Host is not reachable [%s]", asset_id)
        return None

    logging.info("Started inventory discovery for asset [%s]", asset_id)

    os = utils.get_os_release(args, host)
    if os is None:
        logging.error("Failed to identify OS for asset [%s]", asset_id)
        return None

    atype = utils.get_asset_type(os)
    if atype is None:
        logging.error("Could not determine asset type for asset [%s]", asset_id)
        return None

    plist = None
    if atype == 'CentOS' or atype == 'Red Hat' or atype == 'Amazon Linux' or atype == 'Oracle Linux':
        plist = discover_rh(args, host)
    elif atype == 'Ubuntu' or atype == 'Debian':
        plist = discover_ubuntu(args, host)
    elif atype == 'FreeBSD':
        plist = discover_freebsd(args, host)
    elif atype == 'OpenBSD':
        plist = discover_openbsd(args, host)

    if plist == None or len(plist) == 0:
        logging.error("Could not inventory asset [%s]", asset_id)
        return None

    asset_data = {}
    asset_data['id'] = asset_id
    asset_data['name'] = asset_name
    asset_data['type'] = atype
    asset_data['owner'] = args.handle
    asset_data['products'] = plist
    asset_tags = []
    asset_tags.append('OS_RELEASE:' + os)
    asset_tags.append('Linux')
    asset_tags.append(atype)
    asset_data['tags'] = asset_tags

    logging.info("Completed inventory discovery for asset [%s]", asset_id)

    return asset_data

def get_inventory(args):
    return discover(args)
